
#include "db.h"
#include "httpd.h"
#include "socket.h"
#include "malloc.h"
#include "timer.h"

#include <ctype.h>
#include <time.h>

#ifndef _WIN32
#	include <sys/time.h>
#elif defined(_WIN32) && defined(_MSC_VER)
#	include <sys/timeb.h>
#	include <sys/types.h>
#endif

static const char configfile[]="./conf/httpd.conf";	// ʃRtBO

static char logfile[1024]="./log/httpd.log";	// Ot@C
void httpd_set_logfile( const char *str ){ strcpy( logfile, str ); }

static int tz = -1;	// ^C][
void httpd_set_timezone( int tz2 ){ tz=tz2; }

static int max_persist_requests = 32;	// ʐMł̍ő僊NGXg
void httpd_set_max_persist_requests( int i ){ max_persist_requests = i; }

static int request_timeout[] = { 2500, 60*1000 };	// ^CAEg(ŏA)
void httpd_set_request_timeout( int idx, int time ){ request_timeout[idx]=time; }

static char document_root[256]="./httpd/";	// hLg[g
void httpd_set_document_root( const char *str ){ strcpy( document_root,str ); }

enum HTTPD_STATUS {
	HTTPD_REQUEST_WAIT = 0,		// NGXg҂
	HTTPD_REQUEST_WAIT_POST,	// NGXg҂(post)
	HTTPD_REQUEST_OK,			// NGXgߊ
	HTTPD_SEND_HEADER,			// wb_M
	HTTPD_WAITING_SEND			// f[^MI܂ő҂Ă
};

struct httpd_access;

struct httpd_session_data {
	int fd;
	int status;
	int http_ver;
	int header_len;
	int data_len;
	int method;
	int persist;
	int request_count;
	unsigned int tick;
	const unsigned char* url;
	const unsigned char* query;
	unsigned char name[32], passwd[32];
	struct httpd_access *access;
};

enum HTTP_ACCESS {
	HTTPD_ACCESS_ALWAYS		= 0,

	HTTPD_ACCESS_AUTH_NONE	= 0,
	HTTPD_ACCESS_AUTH_BASIC	= 1,
	HTTPD_ACCESS_AUTH_DIGEST= 2,	// T|[g
	HTTPD_ACCESS_AUTH_MASK	= 3,

	HTTPD_ACCESS_IP_NONE	= 0,
	HTTPD_ACCESS_IP_ALLOW	= 4,
	HTTPD_ACCESS_IP_DENY	= 8,
	HTTPD_ACCESS_IP_MUTUAL	=12,
	HTTPD_ACCESS_IP_MASK	=12,

	HTTPD_ACCESS_SAT_ANY	=16,
	HTTPD_ACCESS_SAT_BOTH	= 0,
	HTTPD_ACCESS_SAT_MASK	=16,
};

struct httpd_access_user {
	int type;
	unsigned char name[32];
	unsigned char passwd[32];
};

struct httpd_access {
	unsigned char url[256];
	int type, urllen;

	int dip_count, aip_count, dip_max, aip_max;
	unsigned int *dip, *aip;

	unsigned char realm[256];
	int user_count, user_max;
	struct httpd_access_user *user;
};
static int access_count=0, access_max=0;
static struct httpd_access *access=NULL;

static FILE *logfp=NULL;
void httpd_log( struct httpd_session_data *sd, int status, int len )
{
	static int first=1;
	char timestr[256], lenstr[16];
	unsigned char *ip;
	static const char *method[]={"","GET /","POST/"};
	static const char sign[]={'-','+'};

	// N͂낢Ƃ邱Ƃ
	if( first )
	{
		int tz2;
		first = 0;
		
		// ^C][߂
		{
#ifndef _WIN32
			struct timezone _tz;
			gettimeofday( NULL, &_tz );
			tz2 = _tz.tz_minuteswest;
#elif defined(_MSC_VER)
			struct _timeb _tb;
			_ftime( &_tb );
			tz2 = _tb.timezone;
#else
			tz2=-1;
#endif
		}
		
		if( tz==-1 || tz2==-1)
			tz=tz2;
		else if( tz!=tz2 )
			printf("*notice* : user defined timezone[%d(%d)] != system timezone[%d(%d)]\n", -tz/60, tz, -tz2/60, tz2);
		if( tz==-1 )
		{
			printf("*WARNING* : can't get timezone. plz set by user. temporarily set [+9] (JST)\n");
			tz = -9 * 60;
		}
		
		// Ot@CJ
		if( (logfp = fopen(logfile, "a") ) == 0 )
		{
			printf("*WARNING* : can't open log file [%s]\n", logfile);
		}
	}

	// Ot@CJĂȂ
	if( !logfp )
		return;

	// ԕ߂
#ifdef _WIN32
	{
		time_t time_;
		time(&time_);
		strftime(timestr,sizeof(timestr),"%d/%b/%Y:%H:%M:%S",localtime(&time_) );
	}
#else
	{
		struct timeval tv;

		gettimeofday(&tv,NULL);
		strftime(timestr,sizeof(timestr),"%d/%b/%Y:%H:%M:%S",localtime(&(tv.tv_sec)));
	}
#endif
	sprintf(timestr+20, " %c%02d%02d", sign[(tz<0)?1:0], abs(tz)/60, abs(tz)%60 );
	
	ip = (unsigned char*) &session[sd->fd]->client_addr.sin_addr;
	
	// TCY
	if( len<0 )
	{
		lenstr[0]='-';
		lenstr[1]='\0';
	}
	else
	{
		sprintf(lenstr,"%d",len );
	}
	
	// Oo
	fprintf( logfp,"%d.%d.%d.%d - - [%s] \"%s%s HTTP/1.%d\" %d %s\n",
		ip[0],ip[1],ip[2],ip[3], timestr, method[sd->method], sd->url, sd->http_ver, status, lenstr );
	fflush( logfp );
}


int httpd_strcasencmp(const char *s1, const char *s2,int len) {
	while(len-- && (*s1 || *s2) ) {
		if((*s1 | 0x20) != (*s2 | 0x20)) {
			return ((*s1 | 0x20) > (*s2 | 0x20) ? 1 : -1);
		}
		s1++; s2++;
	}
	return 0;
}

// httpd ɓĂy[WƁAĂяoR[obN֐̈ꗗ
struct dbt* httpd_files;

static int httpd_db_final(void *key,void *data,va_list ap)
{
	char *url = key;

	free(url);

	return 0;
}

static void do_final_httpd(void) {
	int i;
	db_final(httpd_files,httpd_db_final);
	
	for( i=0; i<access_count; i++ )
	{
		aFree( access[i].user );
		aFree( access[i].aip );
		aFree( access[i].dip );
	}
	aFree( access );
	access = NULL;
	access_count = access_max = 0;
}

void do_init_httpd(void) {

//	httpd_config_read(configfile);	// ̊֌WŁAsocket_config_read2() Ă΂

	httpd_files = strdb_init(0);
	atexit(do_final_httpd);
}

// httpd Ƀy[Wǉ
// for ȂǂŃy[Wł悤ɁAkey strdup()̂g

void httpd_pages(const char* url,void (*httpd_func)(struct httpd_session_data*,const char*)) {
	if(strdb_search(httpd_files,url+1) == NULL) {
		strdb_insert(httpd_files,strdup(url+1),httpd_func);
	} else {
		strdb_insert(httpd_files,url+1,httpd_func);
	}
}

static void(*httpd_default)(struct httpd_session_data* sd,const char* url);

void httpd_default_page(void(*httpd_func)(struct httpd_session_data* sd,const char* url)) {
	httpd_default = httpd_func;
}

const char *httpd_get_error( struct httpd_session_data* sd, int* status )
{
	const char* msg;
	// httpd ̃Xe[^X߂
	switch(*status) {
	case 200: msg = "OK";           break;
	case 400: msg = "Bad Request";  break;
	case 401: msg = "Unauthorized"; break;
	case 403: msg = "Forbidden";    break;
	case 404: msg = "Not Found";    break;
	case 408: msg = "Request Timedout"; break;
	case 411: msg = "Length Required"; break;
	case 413: msg = "Request Entity Too Large"; break;
	default:
		*status = 500; msg = "Internal Server Error"; break;
	}
	return msg;
}

void httpd_send_error(struct httpd_session_data* sd,int status) {
	const char* msg = httpd_get_error( sd, &status );
	httpd_send(sd,status,"text/plain",strlen(msg),msg);
}

void httpd_send_head(struct httpd_session_data* sd,int status,const char *content_type,int content_len) {
	char head[1024];
	int  len;
	const char* msg;

	if(sd->status != HTTPD_REQUEST_OK) return;
	msg = httpd_get_error( sd, &status );
	
	// F؊֘Ãf[^Ȃ񂩂
	if( status==401 && ( !sd->access || !(sd->access->type & HTTPD_ACCESS_AUTH_BASIC) ) )
		status = 500;
	
	if(content_len == -1 || ++sd->request_count >= max_persist_requests ) {
		// Ȃ or NGXgE𒴂̂Őؒf
		len = sprintf(
			head,
			"HTTP/1.%d %d %s\r\nContent-Type: %s\r\nConnection: close\r\n",
			sd->http_ver,status,msg,content_type
		);
		sd->persist = 0;
	} else {
		len = sprintf(
			head,
			"HTTP/1.%d %d %s\r\nContent-Type: %s\r\nContent-Length: %d\r\n",
			sd->http_ver,status,msg,content_type,content_len
		);
	}

	if( status==401 )	// F؂Kv
	{
		len += sprintf( head+len,
				"WWW-Authenticate: Basic realm=\"%s\"\r\n", sd->access->realm );
	}
	
	httpd_log( sd, status, content_len );
	
	len += sprintf( head+len, "\r\n" );
	memcpy(WFIFOP(sd->fd,0),head,len);
	WFIFOSET(sd->fd,len);
	sd->status   = HTTPD_SEND_HEADER;
	sd->data_len = content_len;
}

void httpd_send_data(struct httpd_session_data* sd,int content_len,const void *data) {
	const char* msg = (const char*)data;
	if(sd->status == HTTPD_REQUEST_OK) {
		// wb_̑MYĂ̂ŁAKɕ₤
		httpd_send_head(sd,200,"application/octet-stream",-1);
	} else if(sd->status != HTTPD_SEND_HEADER && sd->status != HTTPD_WAITING_SEND) {
		return;
	}
	sd->data_len -= content_len;

	// ȃTCỸt@CMo悤ɕđ
	while(content_len > 0) {
		int send_byte = content_len;
		if(send_byte > 12*1024) send_byte = 12*1024;
		memcpy(WFIFOP(sd->fd,0),msg,send_byte);
		WFIFOSET(sd->fd,send_byte);
		msg += send_byte; content_len -= send_byte;
	}
	sd->status = HTTPD_WAITING_SEND;
}

void httpd_send(struct httpd_session_data* sd,int status,const char *content_type,int content_len,const void *data) {
	httpd_send_head(sd,status,content_type,content_len);
	httpd_send_data(sd,content_len,data);
}

int httpd_disconnect(int fd) {
	// nothing to do
	return 0;
}

void httpd_parse_header(struct httpd_session_data* sd);
void httpd_parse_request_ok(struct httpd_session_data *sd);

int  httpd_parse(int fd) {
	struct httpd_session_data *sd = (struct httpd_session_data*)session[fd]->session_data2;
	if(sd == NULL) {
		sd = (struct httpd_session_data*)aCalloc(sizeof(struct httpd_session_data),1);
		sd->fd  = fd;
		session[fd]->func_destruct = httpd_disconnect;
		session[fd]->session_data2  = sd;
		sd->tick = gettick();
		sd->persist = 0;
		sd->request_count = 0;
	}
	switch(sd->status) {
	case HTTPD_REQUEST_WAIT:
		// NGXg҂
		if(RFIFOREST(fd) > 1024) {
			// NGXĝŁAG[
			sd->status = HTTPD_REQUEST_OK;
			httpd_send_error(sd,400); // Bad Request
		} else if( (int)( gettick() - sd->tick ) > request_timeout[sd->persist] ) {
			// NGXgɎԂ肷Ă̂ŁAG[
			sd->status = HTTPD_REQUEST_OK;
			httpd_send_error(sd,408); // Request Timeout
		} else if(sd->header_len == RFIFOREST(fd)) {
			// ԂȑOƓȂ̂ŁANGXgĉ͂Kv͖
		} else {
			int limit = RFIFOREST(fd);
			unsigned char *req = RFIFOP(fd,0);
			sd->header_len = RFIFOREST(fd);
			do {
				if(*req == '\n' && limit > 0) {
					limit--; req++;
					if(*req == '\r' && limit > 0) { limit--; req++; }
					if(*req == '\n') {
						// HTTPwb_̏I_
						*req   = 0;
						sd->header_len = (req - RFIFOP(fd,0)) + 1;
						httpd_parse_header(sd);
						break;
					}
				}
			} while(req++,--limit > 0);
		}
		break;
	case HTTPD_REQUEST_WAIT_POST:
		if(RFIFOREST(sd->fd) >= sd->header_len) {
			unsigned char temp = RFIFOB(sd->fd,sd->header_len);
			RFIFOB(sd->fd,sd->header_len) = 0;
			httpd_parse_request_ok(sd);
			RFIFOB(sd->fd,sd->header_len) = temp;
		}
		break;
	case HTTPD_REQUEST_OK:
	case HTTPD_SEND_HEADER:
		// NGXgI܂܉MĂȂԂȂ̂ŁA
		// ؒf
		session[fd]->eof = 1;
		break;
	case HTTPD_WAITING_SEND:
		// f[^̑MI܂őҋ@
		if(session[fd]->wdata_size == session[fd]->wdata_pos) {
			// HTTP/1.0͎蓮ؒf
//			if(sd->http_ver == 0) {
			if(sd->persist == 0) {
				session[fd]->eof = 1;
			}
			// RFIFO 烊NGXgf[^̏ƍ\̂̏
			RFIFOSKIP(fd,sd->header_len);
			sd->status     = HTTPD_REQUEST_WAIT;
			sd->tick       = gettick();
			sd->header_len = 0;
			sd->query      = NULL;
//			sd->http_ver   = 0;	// ver ͕ێ
			sd->method     = HTTPD_METHOD_UNKNOWN;
			// printf("httpd_parse: [% 3d] request sended RFIFOREST:%d\n",fd,RFIFOREST(fd));
		}
		break;
	}
	return 0;
}

int httpd_decode_url( char *url )
{
	int s=0, d=0;
	while( url[s]!='\0' && url[s]!='?' )
	{
		if( url[s]=='%' )	// url fR[h
		{
			#define HEX2INT(a) ( ( a>='0' && a<='9' )? a-'0' : (a>='A' && a<='F' )? a-'A'+10 : (a>='a' && a<='f' )? a-'a'+10 : -1 )
			int a1 = HEX2INT(url[s+1]), a2 = HEX2INT(url[s+2]);
			#undef HEX2INT
			if( a1>=2 && a2>=0)	// 0x20`
			{
				s+=3;
				url[ d++ ] = (a1<<4) + a2;
			}
			else if( a1>=0 && a1<2 )	// url ɐ䕶悤Ƃ
			{
				return 0;
			}
			continue;
		}
		url[ d++ ] = url[ s++ ];
	}
	url[d] = '\0';

	// sjis Ɛɂ͂ȂƂ
	s=d=0;
	while( url[s]!='\0' && url[s]!='?' )
	{
		int c0 = url[s], c1 = url[s+1], c2 = url[s+2], c3 = url[s+3];
/*		if( url[s]=='\\' )	// obNNH[g܂܂ĂƒɃG[
		{
			return 0;
		}*/
		
		if( c0=='/' && (c1=='/' || c1=='\\') )	// AXbVFӖ̂ō폜
		{
			s++;
			url[s]='/';
			continue;
		}
		else if( (c0=='/' || c0=='\\') && c1=='.' && (c2=='/' || c2=='\\') )	// JgfBNgFӖȂ̂ō폜
		{
			s+=2;
			url[s]='/';
			continue;
		}
		else if( (c0=='/' || c0=='\\') && c1=='.' && c2=='.' && (c3=='/' || c3=='\\') )	// efBNgFWJ܂̓G[
		{
			int r;
			if( d<=0 )
				return 0;
			for( r=d-1; r>=0 && url[r]!='/'; r-- );
			if( r<0 )
				return 0;
			s+=3;
			d = r;
			url[s]='/';
			continue;
		}
		
		url[ d++ ] = url[ s++ ];
	}
	url[d] = '\0';

	return 1;
}

int httpd_check_access( struct httpd_session_data *sd, int *st )
{
	int i;
	int n=0, len=0;
	struct httpd_access *a;

	// Ԓ}b`T
	for( i=0; i<access_count; i++ )
	{
		struct httpd_access *a = &access[i];
		if( a->urllen > len && httpd_strcasencmp( a->url, sd->url-1, a->urllen )==0 )
			n=i;
	}
	if( i==access_count )
		return 1;
	
	a = &access[n];
		
	if( a->type & HTTPD_ACCESS_IP_MASK )
	{
		int f = 0;
		// todo: IP`FbN
		
		
		if( f && a->type & HTTPD_ACCESS_SAT_ANY )
			return 1;
		else if( !f && !(a->type & HTTPD_ACCESS_SAT_ANY) )
		{
			*st=403;
			return 0;
		}
	}
	else if( a->type & HTTPD_ACCESS_AUTH_MASK )
	{
		int f = 0;
		// todo: [U[`FbN
		
		if( !f )
		{
			*st=401;
			sd->access = a;
			return 0;
		}
	}

	return 1;
}

void httpd_parse_header_sub( struct httpd_session_data *sd, const char *p1, int *plen )
{
	int len = 0;
	// HTTP̃o[W𒲍
	if(!strncmp(p1 ,"HTTP/1.1",8)) {
		sd->http_ver = 1;
		sd->persist  = 1;
	} else {
		sd->http_ver = 0;
		sd->persist  = 0;
	}

	p1  = strchr(p1,'\n');
	while(p1) {
		// Content-Length: ̒
		if(!httpd_strcasencmp(p1+1,"Content-Length: ",16)) {
			len = atoi(p1 + 17);
		}
		// Connection: ̒
		if(!httpd_strcasencmp(p1+1,"Connection: ",12)) {
			if( httpd_strcasencmp(p1+13,"close",5)==0 && sd->http_ver==1 )
				sd->persist = 0;
			if( httpd_strcasencmp(p1+13,"Keep-Alive",10)==0 && sd->http_ver==0 )
				sd->persist = 1;
		}
		p1 = strchr(p1+1,'\n');
	}
	if(plen) *plen = len;
	return;
}

void httpd_parse_header(struct httpd_session_data* sd) {
	int  i;
	int  status = 400; // Bad Request
	unsigned char* req    = RFIFOP(sd->fd,0);
	do {
		if(!strncmp(req,"GET /",5)) {
			// GET NGXg
			req += 5;
			for(i = 0;req[i]; i++) {
				if(req[i] == ' ' || req[i] == '?') break;
				if(	!isalnum(req[i]) && 
					req[i] != '.' && req[i] != '_' && req[i] != '-' &&
					req[i] !='/' && req[i] != '%' )
						break;
			}
			if(req[i] == ' ') {
				req[i]     = 0;
				sd->url    = req;
				sd->query  = NULL;
				sd->status = HTTPD_REQUEST_OK;
			} else if(req[i] == '?') {
				req[i]    = 0;
				sd->query = &req[++i];
				for(;req[i];i++) {
					if(
						isalnum(req[i]) || req[i] == '+' || req[i] == '%' || req[i] == '&' ||
						req[i] == '='
					) {
						continue;
					} else {
						break;
					}
				}
				if(req[i] != ' ') {
					break;
				}
				req[i]     = 0;
				sd->url    = req;
			} else {
				break;
			}
			
			// URL fR[h
			if( !httpd_decode_url( req - 1 ) )
				break;
			
			// wb_
			httpd_parse_header_sub( sd, &req[i+1], NULL );
			
			// ANZX̃`FbN
			if( !httpd_check_access( sd, &status ) )
				break;
			
			// printf("httpd: request %s %s\n",sd->url,sd->query);
			sd->method     = HTTPD_METHOD_GET;
			httpd_parse_request_ok(sd);
		} else if(!strncmp(req,"POST /",6)) {
			int  len;
			req += 6; status = 404;
			for(i = 0;req[i]; i++) {
				if(req[i] == ' ') break;
				if(	!isalnum(req[i]) && 
					req[i] != '.' && req[i] != '_' && req[i] != '-' &&
					req[i] !='/' && req[i] != '%' )
						break;
			}
			if(req[i] != ' ') {
				break;
			}
			req[i]     = 0;
			sd->url    = req;

			// URL fR[h
			if( !httpd_decode_url( req - 1 ) )
				break;
			
			// wb_
			httpd_parse_header_sub( sd, &req[i+1], &len );
			
			// ANZX̃`FbN
			if( !httpd_check_access( sd, &status ) )
				break;
			
			if(len <= 0 || len >= 32*1024) {
				// Ƃ肠32KBȏ̃NGXg͕s
				status = ( len==0 )? 411 : ( len>32*1024 )? 413 : 400;
				break;
			}
			
			sd->query      =  RFIFOP(sd->fd,sd->header_len);
			sd->method     = HTTPD_METHOD_POST;
			sd->header_len += len;
			if(RFIFOREST(sd->fd) >= sd->header_len) {
				unsigned char temp = RFIFOB(sd->fd,sd->header_len);
				RFIFOB(sd->fd,sd->header_len) = 0;
				httpd_parse_request_ok(sd);
				RFIFOB(sd->fd,sd->header_len) = temp;
			} else {
				// POST̃f[^Ă̂҂
				sd->status = HTTPD_REQUEST_WAIT_POST;
			}
		} else {
			break;
		}
	} while(0);
	if(sd->status == HTTPD_REQUEST_WAIT) {
		sd->status = HTTPD_REQUEST_OK;
		httpd_send_error(sd,status); 
	}
}

void httpd_parse_request_ok(struct httpd_session_data *sd) {
	void (*httpd_parse_func)(struct httpd_session_data*,const char*);
	sd->status = HTTPD_REQUEST_OK;

	// t@C܂̂ŁAy[W
	// printf("httpd_parse: [% 3d] request /%s\n",fd,req);
	httpd_parse_func = strdb_search(httpd_files,sd->url);
	if(httpd_parse_func == NULL) {
		httpd_parse_func = httpd_default;
	}
	if(httpd_parse_func == NULL) {
		httpd_send_error(sd,404); // Not Found
	} else {
		httpd_parse_func(sd,sd->url);
		if(sd->status == HTTPD_REQUEST_OK) {
			httpd_send_error(sd,404); // Not Found
		}
	}
	if(sd->persist == 1 && sd->data_len) {
		// ςȃf[^(Ȃ̑Ȃc)
		printf("httpd_parse: send size mismatch when parsing /%s\n",sd->url);
		session[sd->fd]->eof = 1;
	}
	if(sd->status == HTTPD_REQUEST_OK) {
		httpd_send_error(sd,404); 
	}
}

char* httpd_get_value(struct httpd_session_data* sd,const char* val) {
	int src_len = strlen(val);
	const unsigned char* src_p = sd->query;
	if(src_p == NULL) return aStrdup("");

	do {
		if(!memcmp(src_p,val,src_len) && src_p[src_len] == '=') {
			break;
		}
		src_p = strchr(src_p + 1,'&');
		if(src_p) src_p++;
	} while(src_p);

	if(src_p != NULL) {
		// ړI̕
		const unsigned char* p2;
		int   dest_len;
		char* dest_p;
		src_p += src_len + 1;
		p2     = strchr(src_p,'&');
		if(p2 == NULL) {
			src_len = strlen(src_p);
		} else {
			src_len = (p2 - src_p);
		}
		dest_p   = aMalloc(src_len + 1);
		dest_len = 0;
		while(src_len > 0) {
			if(*src_p == '%' && src_len > 2) {
				int c1 = 0,c2 = 0;
				if(src_p[1] >= '0' && src_p[1] <= '9') c1 = src_p[1] - '0';
				if(src_p[1] >= 'A' && src_p[1] <= 'F') c1 = src_p[1] - 'A' + 10;
				if(src_p[1] >= 'a' && src_p[1] <= 'f') c1 = src_p[1] - 'a' + 10;
				if(src_p[2] >= '0' && src_p[2] <= '9') c2 = src_p[2] - '0';
				if(src_p[2] >= 'A' && src_p[2] <= 'F') c2 = src_p[2] - 'A' + 10;
				if(src_p[2] >= 'a' && src_p[2] <= 'f') c2 = src_p[2] - 'a' + 10;
				dest_p[dest_len++] = (c1 << 4) | c2;
				src_p += 3; src_len -= 3;
			} else if(*src_p == '+') {
				dest_p[dest_len++] = ' ';
				src_p++; src_len--;
			} else {
				dest_p[dest_len++] = *(src_p++); src_len--;
			}
		}
		dest_p[dest_len] = 0;
		return dest_p;
	}
	return aStrdup("");
}

int httpd_get_method(struct httpd_session_data* sd) {
	return sd->method;
}

unsigned int httpd_get_ip(struct httpd_session_data *sd) {
	return *(unsigned int*)(&session[sd->fd]->client_addr.sin_addr);
}

// MIME^CvBvȂ̂肵āAcapplication/octet-stream
static const char* httpd_mimetype(const char* url) {
	char *ext = strrchr(url,'.');
	if(ext) {
		if(!strcmp(ext,".html")) return "text/html";
		if(!strcmp(ext,".htm"))  return "text/html";
		if(!strcmp(ext,".css"))  return "text/css";
		if(!strcmp(ext,".js"))   return "text/javascript";
		if(!strcmp(ext,".txt"))  return "text/plain";
		if(!strcmp(ext,".gif"))  return "image/gif";
		if(!strcmp(ext,".jpg"))  return "image/jpeg";
		if(!strcmp(ext,".jpeg")) return "image/jpeg";
		if(!strcmp(ext,".png"))  return "image/png";
		if(!strcmp(ext,".xbm"))  return "image/xbm";
		if(!strcmp(ext,".zip"))  return "application/zip";
	}
	return "application/octet-stream";
}

void httpd_send_file(struct httpd_session_data* sd,const char* url) {
	FILE *fp;
	int  file_size;
	char file_buf[8192];
	if(sd->status != HTTPD_REQUEST_OK) return;

	// URL ꍇ̃ftHg
	if(url[0] == '\0') url = "index.html";
	
	// url ̍ő咷͖1010oCgȂ̂ŁAobt@I[o[t[̐Sz͖
	sprintf(file_buf,"%s%s",document_root,url);

	fp = fopen(file_buf,"rb");
	if(fp == NULL) {
		httpd_send_error(sd,404);
	} else {
		fseek(fp,0,SEEK_END);
		file_size = ftell(fp);
		fseek(fp,0,SEEK_SET);
		httpd_send_head(sd,200,httpd_mimetype(url),file_size);
		while(file_size > 0) {
			int read_byte = file_size;
			if(file_size > 8192) read_byte = 8192;
			fread(file_buf,1,read_byte,fp);
			httpd_send_data(sd,read_byte,file_buf);
			file_size -= read_byte;
		}
		fclose(fp);
	}
}

char* httpd_binary_encode(const char* val) {
	char *buf = aMalloc(strlen(val) * 3 + 1);
	char *p   = buf;
	while(*val) {
		if(isalnum((unsigned char)*val)) {
			*(p++) = *(val++);
		} else {
			unsigned char c1 = *(val++);
			unsigned char c2 = (c1 >> 4);
			unsigned char c3 = (c1 & 0x0F);
			*(p++) = '%';
			*(p++) = c2 + (c2 >= 10 ? 'A'-10 : '0');
			*(p++) = c3 + (c3 >= 10 ? 'A'-10 : '0');
		}
	}
	*p = 0;
	return buf;
}

char* httpd_quote_meta(const char* p1) {
	char *buf = aMalloc(strlen(p1) * 6 + 1);
	char *p2  = buf;
	while(*p1) {
		switch(*p1) {
		case '<': memcpy(p2,"&lt;",4);   p2 += 4; p1++; break;
		case '>': memcpy(p2,"&gt;",4);   p2 += 4; p1++; break;
		case '&': memcpy(p2,"&amp;",5);  p2 += 5; p1++; break;
		case '"': memcpy(p2,"&quot;",6); p2 += 6; p1++; break;
		default: *(p2++) = *(p1++);
		}
	}
	*p2 = 0;
	return buf;
}


void httpd_config_read_add_authuser( struct httpd_access *a, const char *name, const char *passwd )
{
	struct httpd_access_user* au;
	if( a->user_count == a->user_max )
	{
		struct httpd_access_user* au = aMalloc( sizeof(struct httpd_access_user) * (a->user_max + 16) );
		if( a->user )
		{
			memcpy( au, a->user, sizeof(struct httpd_access_user) * a->user_count );
			aFree( a->user );
		}
		a->user = au;
		a->user_max += 16;
	}
	au = a->user + (a->user_count++);
	au->type = 0;
	strcpy( au->name, name );
	strcpy( au->passwd, passwd );
}

void httpd_config_read_add_ip( unsigned int **list, int *count, int *max, const char *w2 )
{
	int i1,i2,i3,i4, m1,m2,m3,m4;
	unsigned int ip, mask;
	if( sscanf( w2, "%d.%d.%d.%d/%d.%d.%d.%d", &i1, &i2, &i3, &i4, &m1, &m2, &m3, &m4 )==8 )
	{
		ip   = (i1<<24)|(i2<<16)|(i3<<8)|i4;
		mask = (m1<<24)|(m2<<16)|(m3<<8)|m4;
	}
	else if( sscanf( w2,"%d.%d.%d.%d/%d", &i1, &i2, &i3, &i4, &m1 )==5 )
	{
		ip   = (i1<<24)|(i2<<16)|(i3<<8)|i4;
		mask = 0xffffffff << (32-m1);
	}
	else if( sscanf( w2,"%d.%d.%d.%d", &i1, &i2, &i3, &i4 )==4 )
	{
		ip   = (i1<<24)|(i2<<16)|(i3<<8)|i4;
		mask = 0xffffffff;
	}
	else if( sscanf( w2,"%d.%d.%d", &i1, &i2, &i3 )==3 )
	{
		ip   = (i1<<24)|(i2<<16)|(i3<<8);
		mask = 0xffffff00;
	}
	else if( sscanf( w2,"%d.%d", &i1, &i2 )==2 )
	{
		ip   = (i1<<24)|(i2<<16);
		mask = 0xffff0000;
	}
	else
	{
		ip   = atoi(w2)<<24;
		mask = 0xff000000;
	}

	if( *count == *max )
	{
		unsigned int *iplist = aMalloc( sizeof(unsigned int) * (*max + 16) );
		if( *list )
		{
			memcpy( iplist, *list, sizeof(unsigned int) * (*count) );
			aFree( *list );
		}
		*list = iplist;
		*max += 16;
	}
	(*list)[ (*count)++ ] = ip;
	(*list)[ (*count)++ ] = mask;
}

/*==========================================
 * ݒt@Cǂݍ
 *------------------------------------------
 */
int httpd_config_read(char *cfgName)
{
	int i;
	char line[1024],w1[1024],w2[1024];
	FILE *fp;
	struct httpd_access *a = NULL;

	fp=fopen(cfgName,"r");
	if(fp==NULL){
		printf("file not found: %s\n",cfgName);
		return 1;
	}
	while(fgets(line,1020,fp)){
		if(line[0] == '/' && line[1] == '/')
			continue;
		i=sscanf(line,"%[^:]: %[^\r\n]",w1,w2);
		if(i!=2)
			continue;

		if(strcmpi(w1,"enable")==0)
		{
			socket_enable_httpd( atoi(w2) );
		}
		else if(strcmpi(w1,"log_filename")==0)
		{
			strcpy( logfile, w2 );
		}
		else if(strcmpi(w1,"request_timeout_first")==0)
		{
			httpd_set_request_timeout( 0, atoi(w2) );
		}
		else if(strcmpi(w1,"request_timeout_persist")==0)
		{
			httpd_set_request_timeout( 1, atoi(w2) );
		}
		else if(strcmpi(w1,"max_persist_requests")==0)
		{
			httpd_set_max_persist_requests( atoi(w2) );
		}
		else if(strcmpi(w1,"timezone")==0)
		{
			if( strcmpi(w2,"auto") == 0)
				httpd_set_timezone( -1 );
			else
				httpd_set_timezone( atoi(w2)*(-60) );
		}
		else if(strcmpi(w1,"document_root")==0)
		{
			httpd_set_document_root( w2 );
		}
		else if( strcmpi(w1,"target")==0 )
		{
			if( access_count==access_max )
			{
				struct httpd_access *a = aMalloc( sizeof( struct httpd_access ) * access_max+16 );
				if( access )
				{
					memcpy( a, access, sizeof( struct httpd_access ) * access_count );
					aFree( access );
				}
				access = a;
				access_max += 16;
			}
			a = access+(access_count++);
			a->type = HTTPD_ACCESS_ALWAYS;
			a->aip_count = a->aip_max = 0;
			a->dip_count = a->dip_max = 0;
			a->user_count= a->user_max= 0;
			a->aip = a->dip = NULL;
			a->user = NULL;
			a->url[0]= '\0';
			a->urllen = 0;
			if( strlen(w2)>=sizeof(a->url) )
				printf("httpd_config_read: too long target url [%s]\n", w2);
			else
			{
				strcpy( a->url, w2 );
				a->urllen = strlen(a->url);
			}
		}
		else if(strcmpi(w1,"satisfy")==0 )
		{
			int i = (strcmpi(w2,"any")==0)? HTTPD_ACCESS_SAT_ANY : (strcmpi(w2,"both")==0)? HTTPD_ACCESS_SAT_BOTH : -1;
			if( i<0 )
				printf("httpd_config_read: satisfy: unknown satisfy option [%s]\n", w2);
			else
				a->type = (a->type & ~HTTPD_ACCESS_SAT_MASK) | i;
		}
		else if(strcmpi(w1,"authtype")==0)
		{
			int i = (strcmpi(w2,"basic")==0)? HTTPD_ACCESS_AUTH_BASIC : -1;
			if( i<0 )
				printf("httpd_config_read: authtype: unknown authtype [%s]\n", w2);
			else
				a->type = (a->type & ~HTTPD_ACCESS_AUTH_MASK) | i;
		}
		else if(strcmpi(w1,"authname")==0 || strcmpi(w1,"authrealm")==0 )
		{
			if( strlen(w2)>=sizeof(a->realm) )
				printf("httpd_config_read: %s: too long realm [%s]\n", w1, w2);
			else
				strcpy( a->realm, w2 );
		}
		else if(strcmpi(w1,"authuser")==0 )
		{
			char u[1024], p[1024];
			if( sscanf(w2,"%[^:]:%s",u,p) == 2 )
			{
				if( strlen(u)>=sizeof(a->user->name) || strlen(p)>=sizeof(a->user->passwd) )
					printf("httpd_config_read: authuser: too long name or passwd\n");
				else
					httpd_config_read_add_authuser( a, u, p );
			}
			else	// Ot@Cǂݍ߂肷Ƃ
			{
				printf("httpd_config_read: authuser: [user:pass] needed\n");
			}
		}
		else if(strcmpi(w1,"order")==0 )
		{
			int i = (strcmpi(w2,"deny,allow"    )==0)? HTTPD_ACCESS_IP_DENY   :
					(strcmpi(w2,"allow,deny"    )==0)? HTTPD_ACCESS_IP_ALLOW  :
					(strcmpi(w2,"mutual-failure")==0)? HTTPD_ACCESS_IP_MUTUAL : -1;
			if( i<0 )
				printf("httpd_config_read: order: unknown order option [%s]\n", w2);
			else
				a->type = (a->type & ~HTTPD_ACCESS_IP_MASK) | i;
		}
		else if(strcmpi(w1,"allow")==0 )
		{
			httpd_config_read_add_ip( &a->aip, &a->aip_count, &a->aip_max, w2 );
		}
		else if(strcmpi(w1,"deny")==0 )
		{
			httpd_config_read_add_ip( &a->aip, &a->aip_count, &a->aip_max, w2 );
		}
		else if(strcmpi(w1,"import")==0)
		{
			httpd_config_read(w2);
		}
		else
		{
			printf("httpd_config_read: unknown option [%s: %s]",w1,w2);
		}
	}
	fclose(fp);

	return 0;
}

